@@ -28,20 +28,23 @@
 #include "replace.h"
 #include "system/filesys.h"
 
-#ifndef Py_RETURN_NONE
-#define Py_RETURN_NONE return Py_INCREF(Py_None), Py_None
-#endif
-
 /* Include tdb headers */
 #include <tdb.h>
 
+#if PY_MAJOR_VERSION >= 3
+#define Py_TPFLAGS_HAVE_ITER 0
+#endif
+
+/* discard signature of 'func' in favour of 'target_sig' */
+#define PY_DISCARD_FUNC_SIG(target_sig, func) (target_sig)(void(*)(void))func
+
 typedef struct {
 	PyObject_HEAD
 	TDB_CONTEXT *ctx;
 	bool closed;
 } PyTdbObject;
 
-staticforward PyTypeObject PyTdb;
+static PyTypeObject PyTdb;
 
 static void PyErr_SetTDBError(TDB_CONTEXT *tdb)
 {
@@ -49,21 +52,21 @@ static void PyErr_SetTDBError(TDB_CONTEXT *tdb)
 		Py_BuildValue("(i,s)", tdb_error(tdb), tdb_errorstr(tdb)));
 }
 
-static TDB_DATA PyString_AsTDB_DATA(PyObject *data)
+static TDB_DATA PyBytes_AsTDB_DATA(PyObject *data)
 {
 	TDB_DATA ret;
-	ret.dptr = (unsigned char *)PyString_AsString(data);
-	ret.dsize = PyString_Size(data);
+	ret.dptr = (unsigned char *)PyBytes_AsString(data);
+	ret.dsize = PyBytes_Size(data);
 	return ret;
 }
 
-static PyObject *PyString_FromTDB_DATA(TDB_DATA data)
+static PyObject *PyBytes_FromTDB_DATA(TDB_DATA data)
 {
 	if (data.dptr == NULL && data.dsize == 0) {
 		Py_RETURN_NONE;
 	} else {
-		PyObject *ret = PyString_FromStringAndSize((const char *)data.dptr, 
-												   data.dsize);
+		PyObject *ret = PyBytes_FromStringAndSize((const char *)data.dptr,
+												  data.dsize);
 		free(data.dptr);
 		return ret;
     }
@@ -95,9 +98,10 @@ static PyObject *py_tdb_open(PyTypeObject *type, PyObject *args, PyObject *kwarg
 	int hash_size = 0, tdb_flags = TDB_DEFAULT, flags = O_RDWR, mode = 0600;
 	TDB_CONTEXT *ctx;
 	PyTdbObject *ret;
-	const char *kwnames[] = { "name", "hash_size", "tdb_flags", "flags", "mode", NULL };
+	const char *_kwnames[] = { "name", "hash_size", "tdb_flags", "flags", "mode", NULL };
+	char **kwnames = discard_const_p(char *, _kwnames);
 
-	if (!PyArg_ParseTupleAndKeywords(args, kwargs, "|siiii", (char **)kwnames, &name, &hash_size, &tdb_flags, &flags, &mode))
+	if (!PyArg_ParseTupleAndKeywords(args, kwargs, "|siiii", kwnames, &name, &hash_size, &tdb_flags, &flags, &mode))
 		return NULL;
 
 	if (name == NULL) {
@@ -121,7 +125,8 @@ static PyObject *py_tdb_open(PyTypeObject *type, PyObject *args, PyObject *kwarg
 	return (PyObject *)ret;
 }
 
-static PyObject *obj_transaction_cancel(PyTdbObject *self)
+static PyObject *obj_transaction_cancel(PyTdbObject *self,
+		PyObject *Py_UNUSED(ignored))
 {
 	int ret;
 
@@ -132,7 +137,8 @@ static PyObject *obj_transaction_cancel(PyTdbObject *self)
 	Py_RETURN_NONE;
 }
 
-static PyObject *obj_transaction_commit(PyTdbObject *self)
+static PyObject *obj_transaction_commit(PyTdbObject *self,
+		PyObject *Py_UNUSED(ignored))
 {
 	int ret;
 	PyErr_TDB_RAISE_IF_CLOSED(self);
@@ -141,7 +147,8 @@ static PyObject *obj_transaction_commit(PyTdbObject *self)
 	Py_RETURN_NONE;
 }
 
-static PyObject *obj_transaction_prepare_commit(PyTdbObject *self)
+static PyObject *obj_transaction_prepare_commit(PyTdbObject *self,
+		PyObject *Py_UNUSED(ignored))
 {
 	int ret;
 	PyErr_TDB_RAISE_IF_CLOSED(self);
@@ -150,7 +157,8 @@ static PyObject *obj_transaction_prepare_commit(PyTdbObject *self)
 	Py_RETURN_NONE;
 }
 
-static PyObject *obj_transaction_start(PyTdbObject *self)
+static PyObject *obj_transaction_start(PyTdbObject *self,
+		PyObject *Py_UNUSED(ignored))
 {
 	int ret;
 	PyErr_TDB_RAISE_IF_CLOSED(self);
@@ -159,7 +167,8 @@ static PyObject *obj_transaction_start(PyTdbObject *self)
 	Py_RETURN_NONE;
 }
 
-static PyObject *obj_reopen(PyTdbObject *self)
+static PyObject *obj_reopen(PyTdbObject *self,
+		PyObject *Py_UNUSED(ignored))
 {
 	int ret;
 	PyErr_TDB_RAISE_IF_CLOSED(self);
@@ -175,7 +184,8 @@ static PyObject *obj_reopen(PyTdbObject *self)
 	Py_RETURN_NONE;
 }
 
-static PyObject *obj_lockall(PyTdbObject *self)
+static PyObject *obj_lockall(PyTdbObject *self,
+		PyObject *Py_UNUSED(ignored))
 {
 	int ret;
 	PyErr_TDB_RAISE_IF_CLOSED(self);
@@ -184,7 +194,8 @@ static PyObject *obj_lockall(PyTdbObject *self)
 	Py_RETURN_NONE;
 }
 
-static PyObject *obj_unlockall(PyTdbObject *self)
+static PyObject *obj_unlockall(PyTdbObject *self,
+		PyObject *Py_UNUSED(ignored))
 {
 	int ret;
 	PyErr_TDB_RAISE_IF_CLOSED(self);
@@ -193,7 +204,8 @@ static PyObject *obj_unlockall(PyTdbObject *self)
 	Py_RETURN_NONE;
 }
 
-static PyObject *obj_lockall_read(PyTdbObject *self)
+static PyObject *obj_lockall_read(PyTdbObject *self,
+		PyObject *Py_UNUSED(ignored))
 {
 	int ret;
 	PyErr_TDB_RAISE_IF_CLOSED(self);
@@ -202,14 +214,15 @@ static PyObject *obj_lockall_read(PyTdbObject *self)
 	Py_RETURN_NONE;
 }
 
-static PyObject *obj_unlockall_read(PyTdbObject *self)
+static PyObject *obj_unlockall_read(PyTdbObject *self,
+		PyObject *Py_UNUSED(ignored))
 {
 	int ret = tdb_unlockall_read(self->ctx);
 	PyErr_TDB_ERROR_IS_ERR_RAISE(ret, self->ctx);
 	Py_RETURN_NONE;
 }
 
-static PyObject *obj_close(PyTdbObject *self)
+static PyObject *obj_close(PyTdbObject *self, PyObject *Py_UNUSED(ignored))
 {
 	int ret;
 	if (self->closed)
@@ -236,11 +249,11 @@ static PyObject *obj_get(PyTdbObject *self, PyObject *args)
 	if (!PyArg_ParseTuple(args, "O", &py_key))
 		return NULL;
 
-	key = PyString_AsTDB_DATA(py_key);
+	key = PyBytes_AsTDB_DATA(py_key);
 	if (!key.dptr)
 		return NULL;
 
-	return PyString_FromTDB_DATA(tdb_fetch(self->ctx, key));
+	return PyBytes_FromTDB_DATA(tdb_fetch(self->ctx, key));
 }
 
 static PyObject *obj_append(PyTdbObject *self, PyObject *args)
@@ -254,10 +267,10 @@ static PyObject *obj_append(PyTdbObject *self, PyObject *args)
 	if (!PyArg_ParseTuple(args, "OO", &py_key, &py_data))
 		return NULL;
 
-	key = PyString_AsTDB_DATA(py_key);
+	key = PyBytes_AsTDB_DATA(py_key);
 	if (!key.dptr)
 		return NULL;
-	data = PyString_AsTDB_DATA(py_data);
+	data = PyBytes_AsTDB_DATA(py_data);
 	if (!data.dptr)
 		return NULL;
 
@@ -266,11 +279,11 @@ static PyObject *obj_append(PyTdbObject *self, PyObject *args)
 	Py_RETURN_NONE;
 }
 
-static PyObject *obj_firstkey(PyTdbObject *self)
+static PyObject *obj_firstkey(PyTdbObject *self, PyObject *Py_UNUSED(ignored))
 {
 	PyErr_TDB_RAISE_IF_CLOSED(self);
 
-	return PyString_FromTDB_DATA(tdb_firstkey(self->ctx));
+	return PyBytes_FromTDB_DATA(tdb_firstkey(self->ctx));
 }
 
 static PyObject *obj_nextkey(PyTdbObject *self, PyObject *args)
@@ -282,11 +295,11 @@ static PyObject *obj_nextkey(PyTdbObject *self, PyObject *args)
 	if (!PyArg_ParseTuple(args, "O", &py_key))
 		return NULL;
 
-	key = PyString_AsTDB_DATA(py_key);
+	key = PyBytes_AsTDB_DATA(py_key);
 	if (!key.dptr)
 		return NULL;
 	
-	return PyString_FromTDB_DATA(tdb_nextkey(self->ctx, key));
+	return PyBytes_FromTDB_DATA(tdb_nextkey(self->ctx, key));
 }
 
 static PyObject *obj_delete(PyTdbObject *self, PyObject *args)
@@ -299,7 +312,7 @@ static PyObject *obj_delete(PyTdbObject *self, PyObject *args)
 	if (!PyArg_ParseTuple(args, "O", &py_key))
 		return NULL;
 
-	key = PyString_AsTDB_DATA(py_key);
+	key = PyBytes_AsTDB_DATA(py_key);
 	if (!key.dptr)
 		return NULL;
 	ret = tdb_delete(self->ctx, key);
@@ -307,9 +320,26 @@ static PyObject *obj_delete(PyTdbObject *self, PyObject *args)
 	Py_RETURN_NONE;
 }
 
-static PyObject *obj_has_key(PyTdbObject *self, PyObject *args)
+static int obj_contains(PyTdbObject *self, PyObject *py_key)
 {
 	TDB_DATA key;
+	int ret;
+	PyErr_TDB_RAISE_RETURN_MINUS_1_IF_CLOSED(self);
+
+	key = PyBytes_AsTDB_DATA(py_key);
+	if (!key.dptr) {
+		PyErr_BadArgument();
+		return -1;
+	}
+	ret = tdb_exists(self->ctx, key);
+	if (ret)
+		return 1;
+	return 0;
+}
+
+#if PY_MAJOR_VERSION < 3
+static PyObject *obj_has_key(PyTdbObject *self, PyObject *args)
+{
 	int ret;
 	PyObject *py_key;
 	PyErr_TDB_RAISE_IF_CLOSED(self);
@@ -317,16 +347,15 @@ static PyObject *obj_has_key(PyTdbObject *self, PyObject *args)
 	if (!PyArg_ParseTuple(args, "O", &py_key))
 		return NULL;
 
-	key = PyString_AsTDB_DATA(py_key);
-	if (!key.dptr)
+	ret = obj_contains(self, py_key);
+	if (ret == -1)
 		return NULL;
-	ret = tdb_exists(self->ctx, key);
-	if (ret != TDB_ERR_NOEXIST) {
-		PyErr_TDB_ERROR_IS_ERR_RAISE(ret, self->ctx);
-	}
+	if (ret)
+		Py_RETURN_TRUE;
+	Py_RETURN_FALSE;
 
-	return (ret == TDB_ERR_NOEXIST)?Py_False:Py_True;
 }
+#endif
 
 static PyObject *obj_store(PyTdbObject *self, PyObject *args)
 {
@@ -340,10 +369,10 @@ static PyObject *obj_store(PyTdbObject *self, PyObject *args)
 	if (!PyArg_ParseTuple(args, "OO|i", &py_key, &py_value, &flag))
 		return NULL;
 
-	key = PyString_AsTDB_DATA(py_key);
+	key = PyBytes_AsTDB_DATA(py_key);
 	if (!key.dptr)
 		return NULL;
-	value = PyString_AsTDB_DATA(py_value);
+	value = PyBytes_AsTDB_DATA(py_value);
 	if (!value.dptr)
 		return NULL;
 
@@ -352,6 +381,53 @@ static PyObject *obj_store(PyTdbObject *self, PyObject *args)
 	Py_RETURN_NONE;
 }
 
+static PyObject *obj_storev(PyTdbObject *self, PyObject *args)
+{
+	TDB_DATA key, *values, value;
+	int ret;
+	int flag = TDB_REPLACE;
+	Py_ssize_t num_values, i;
+	PyObject *py_key, *py_values, *py_value;
+
+	PyErr_TDB_RAISE_IF_CLOSED(self);
+
+	if (!PyArg_ParseTuple(
+		    args, "OO!|i", &py_key, &PyList_Type, &py_values, &flag)) {
+		return NULL;
+	}
+
+	num_values = PyList_Size(py_values);
+
+	key = PyBytes_AsTDB_DATA(py_key);
+	if (key.dptr == NULL) {
+		return NULL;
+	}
+
+	if (SSIZE_MAX/sizeof(TDB_DATA) < num_values) {
+		PyErr_SetFromErrno(PyExc_OverflowError);
+		return NULL;
+	}
+	values = malloc(sizeof(TDB_DATA) * num_values);
+	if (values == NULL) {
+		PyErr_NoMemory();
+		return NULL;
+	}
+	for (i=0; i<num_values; i++) {
+		py_value = PyList_GetItem(py_values, i);
+		value = PyBytes_AsTDB_DATA(py_value);
+		if (!value.dptr) {
+			free(values);
+			return NULL;
+		}
+		values[i] = value;
+	}
+
+	ret = tdb_storev(self->ctx, key, values, num_values, flag);
+	free(values);
+	PyErr_TDB_ERROR_IS_ERR_RAISE(ret, self->ctx);
+	Py_RETURN_NONE;
+}
+
 static PyObject *obj_add_flags(PyTdbObject *self, PyObject *args)
 {
 	unsigned flags;
@@ -392,7 +468,7 @@ static PyObject *tdb_iter_next(PyTdbIteratorObject *self)
 		return NULL;
 	current = self->current;
 	self->current = tdb_nextkey(self->iteratee->ctx, self->current);
-	ret = PyString_FromTDB_DATA(current);
+	ret = PyBytes_FromTDB_DATA(current);
 	return ret;
 }
 
@@ -411,7 +487,8 @@ PyTypeObject PyTdbIterator = {
 	.tp_iter = PyObject_SelfIter,
 };
 
-static PyObject *tdb_object_iter(PyTdbObject *self)
+static PyObject *tdb_object_iter(PyTdbObject *self,
+		PyObject *Py_UNUSED(ignored))
 {
 	PyTdbIteratorObject *ret;	
 
@@ -426,7 +503,7 @@ static PyObject *tdb_object_iter(PyTdbObject *self)
 	return (PyObject *)ret;
 }
 
-static PyObject *obj_clear(PyTdbObject *self)
+static PyObject *obj_clear(PyTdbObject *self, PyObject *Py_UNUSED(ignored))
 {
 	int ret;
 	PyErr_TDB_RAISE_IF_CLOSED(self);
@@ -435,7 +512,7 @@ static PyObject *obj_clear(PyTdbObject *self)
 	Py_RETURN_NONE;
 }
 
-static PyObject *obj_repack(PyTdbObject *self)
+static PyObject *obj_repack(PyTdbObject *self, PyObject *Py_UNUSED(ignored))
 {
 	int ret;
 	PyErr_TDB_RAISE_IF_CLOSED(self);
@@ -444,14 +521,16 @@ static PyObject *obj_repack(PyTdbObject *self)
 	Py_RETURN_NONE;
 }
 
-static PyObject *obj_enable_seqnum(PyTdbObject *self)
+static PyObject *obj_enable_seqnum(PyTdbObject *self,
+		PyObject *Py_UNUSED(ignored))
 {
 	PyErr_TDB_RAISE_IF_CLOSED(self);
 	tdb_enable_seqnum(self->ctx);
 	Py_RETURN_NONE;
 }
 
-static PyObject *obj_increment_seqnum_nonblock(PyTdbObject *self)
+static PyObject *obj_increment_seqnum_nonblock(PyTdbObject *self,
+		PyObject *Py_UNUSED(ignored))
 {
 	PyErr_TDB_RAISE_IF_CLOSED(self);
 	tdb_increment_seqnum_nonblock(self->ctx);
@@ -483,17 +562,25 @@ static PyMethodDef tdb_object_methods[] = {
 		"Append data to an existing key." },
 	{ "firstkey", (PyCFunction)obj_firstkey, METH_NOARGS, "S.firstkey() -> data\n"
 		"Return the first key in this database." },
-	{ "nextkey", (PyCFunction)obj_nextkey, METH_NOARGS, "S.nextkey(key) -> data\n"
+	{ "nextkey", (PyCFunction)obj_nextkey, METH_VARARGS, "S.nextkey(key) -> data\n"
 		"Return the next key in this database." },
 	{ "delete", (PyCFunction)obj_delete, METH_VARARGS, "S.delete(key) -> None\n"
 		"Delete an entry." },
+#if PY_MAJOR_VERSION < 3
 	{ "has_key", (PyCFunction)obj_has_key, METH_VARARGS, "S.has_key(key) -> None\n"
 		"Check whether key exists in this database." },
+#endif
 	{ "store", (PyCFunction)obj_store, METH_VARARGS, "S.store(key, data, flag=REPLACE) -> None"
 		"Store data." },
+	{ "storev", (PyCFunction)obj_storev, METH_VARARGS, "S.storev(key, data, flag=REPLACE) -> None"
+		"Store several data." },
 	{ "add_flags", (PyCFunction)obj_add_flags, METH_VARARGS, "S.add_flags(flags) -> None" },
 	{ "remove_flags", (PyCFunction)obj_remove_flags, METH_VARARGS, "S.remove_flags(flags) -> None" },
+#if PY_MAJOR_VERSION >= 3
+	{ "keys", (PyCFunction)tdb_object_iter, METH_NOARGS, "S.iterkeys() -> iterator" },
+#else
 	{ "iterkeys", (PyCFunction)tdb_object_iter, METH_NOARGS, "S.iterkeys() -> iterator" },
+#endif
 	{ "clear", (PyCFunction)obj_clear, METH_NOARGS, "S.clear() -> None\n"
 		"Wipe the entire database." },
 	{ "repack", (PyCFunction)obj_repack, METH_NOARGS, "S.repack() -> None\n"
@@ -502,73 +589,115 @@ static PyMethodDef tdb_object_methods[] = {
 		"S.enable_seqnum() -> None" },
 	{ "increment_seqnum_nonblock", (PyCFunction)obj_increment_seqnum_nonblock, METH_NOARGS,
 		"S.increment_seqnum_nonblock() -> None" },
-	{ NULL }
+	{0}
 };
 
 static PyObject *obj_get_hash_size(PyTdbObject *self, void *closure)
 {
 	PyErr_TDB_RAISE_IF_CLOSED(self);
-	return PyInt_FromLong(tdb_hash_size(self->ctx));
+	return PyLong_FromLong(tdb_hash_size(self->ctx));
 }
 
 static int obj_set_max_dead(PyTdbObject *self, PyObject *max_dead, void *closure)
 {
 	PyErr_TDB_RAISE_RETURN_MINUS_1_IF_CLOSED(self);
-	if (!PyInt_Check(max_dead))
+	if (!PyLong_Check(max_dead))
 		return -1;
-	tdb_set_max_dead(self->ctx, PyInt_AsLong(max_dead));
+	tdb_set_max_dead(self->ctx, PyLong_AsLong(max_dead));
 	return 0;
 }
 
 static PyObject *obj_get_map_size(PyTdbObject *self, void *closure)
 {
 	PyErr_TDB_RAISE_IF_CLOSED(self);
-	return PyInt_FromLong(tdb_map_size(self->ctx));
+	return PyLong_FromLong(tdb_map_size(self->ctx));
 }
 
 static PyObject *obj_get_freelist_size(PyTdbObject *self, void *closure)
 {
 	PyErr_TDB_RAISE_IF_CLOSED(self);
-	return PyInt_FromLong(tdb_freelist_size(self->ctx));
+	return PyLong_FromLong(tdb_freelist_size(self->ctx));
 }
 
 static PyObject *obj_get_flags(PyTdbObject *self, void *closure)
 {
 	PyErr_TDB_RAISE_IF_CLOSED(self);
-	return PyInt_FromLong(tdb_get_flags(self->ctx));
+	return PyLong_FromLong(tdb_get_flags(self->ctx));
 }
 
 static PyObject *obj_get_filename(PyTdbObject *self, void *closure)
 {
 	PyErr_TDB_RAISE_IF_CLOSED(self);
-	return PyString_FromString(tdb_name(self->ctx));
+	return PyBytes_FromString(tdb_name(self->ctx));
 }
 
 static PyObject *obj_get_seqnum(PyTdbObject *self, void *closure)
 {
 	PyErr_TDB_RAISE_IF_CLOSED(self);
-	return PyInt_FromLong(tdb_get_seqnum(self->ctx));
+	return PyLong_FromLong(tdb_get_seqnum(self->ctx));
 }
 
+static PyObject *obj_get_text(PyTdbObject *self, void *closure)
+{
+	PyObject *mod, *cls, *inst;
+	mod = PyImport_ImportModule("_tdb_text");
+	if (mod == NULL)
+		return NULL;
+	cls = PyObject_GetAttrString(mod, "TdbTextWrapper");
+	if (cls == NULL) {
+		Py_DECREF(mod);
+		return NULL;
+	}
+	inst = PyObject_CallFunction(cls, discard_const_p(char, "O"), self);
+	Py_DECREF(mod);
+	Py_DECREF(cls);
+	return inst;
+}
 
 static PyGetSetDef tdb_object_getsetters[] = {
-	{ (char *)"hash_size", (getter)obj_get_hash_size, NULL, NULL },
-	{ (char *)"map_size", (getter)obj_get_map_size, NULL, NULL },
-	{ (char *)"freelist_size", (getter)obj_get_freelist_size, NULL, NULL },
-	{ (char *)"flags", (getter)obj_get_flags, NULL, NULL },
-	{ (char *)"max_dead", NULL, (setter)obj_set_max_dead, NULL },
-	{ (char *)"filename", (getter)obj_get_filename, NULL, (char *)"The filename of this TDB file."},
-	{ (char *)"seqnum", (getter)obj_get_seqnum, NULL, NULL },
-	{ NULL }
+	{
+		.name    = discard_const_p(char, "hash_size"),
+		.get     = (getter)obj_get_hash_size,
+	},
+	{
+		.name    = discard_const_p(char, "map_size"),
+		.get     = (getter)obj_get_map_size,
+	},
+	{
+		.name    = discard_const_p(char, "freelist_size"),
+		.get     = (getter)obj_get_freelist_size,
+	},
+	{
+		.name    = discard_const_p(char, "flags"),
+		.get     = (getter)obj_get_flags,
+	},
+	{
+		.name    = discard_const_p(char, "max_dead"),
+		.set     = (setter)obj_set_max_dead,
+	},
+	{
+		.name    = discard_const_p(char, "filename"),
+		.get     = (getter)obj_get_filename,
+		.doc     = discard_const_p(char, "The filename of this TDB file."),
+	},
+	{
+		.name    = discard_const_p(char, "seqnum"),
+		.get     = (getter)obj_get_seqnum,
+	},
+	{
+		.name    = discard_const_p(char, "text"),
+		.get     = (getter)obj_get_text,
+	},
+	{ .name = NULL }
 };
 
 static PyObject *tdb_object_repr(PyTdbObject *self)
 {
 	PyErr_TDB_RAISE_IF_CLOSED(self);
 	if (tdb_get_flags(self->ctx) & TDB_INTERNAL) {
-		return PyString_FromString("Tdb(<internal>)");
+		return PyUnicode_FromString("Tdb(<internal>)");
 	} else {
-		return PyString_FromFormat("Tdb('%s')", tdb_name(self->ctx));
+		return PyUnicode_FromFormat("Tdb('%s')", tdb_name(self->ctx));
 	}
 }
 
@@ -576,27 +705,31 @@ static void tdb_object_dealloc(PyTdbObject *self)
 {
 	if (!self->closed)
 		tdb_close(self->ctx);
-	self->ob_type->tp_free(self);
+	Py_TYPE(self)->tp_free(self);
 }
 
 static PyObject *obj_getitem(PyTdbObject *self, PyObject *key)
 {
 	TDB_DATA tkey, val;
 	PyErr_TDB_RAISE_IF_CLOSED(self);
-	if (!PyString_Check(key)) {
-		PyErr_SetString(PyExc_TypeError, "Expected string as key");
+	if (!PyBytes_Check(key)) {
+		PyErr_SetString(PyExc_TypeError, "Expected bytestring as key");
 		return NULL;
 	}
 
-	tkey.dptr = (unsigned char *)PyString_AsString(key);
-	tkey.dsize = PyString_Size(key);
+	tkey.dptr = (unsigned char *)PyBytes_AsString(key);
+	tkey.dsize = PyBytes_Size(key);
 
 	val = tdb_fetch(self->ctx, tkey);
 	if (val.dptr == NULL) {
-		PyErr_SetString(PyExc_KeyError, "No such TDB entry");
+		/*
+		 * if the key doesn't exist raise KeyError(key) to be
+		 * consistent with python dict
+		 */
+		PyErr_SetObject(PyExc_KeyError, key);
 		return NULL;
 	} else {
-		return PyString_FromTDB_DATA(val);
+		return PyBytes_FromTDB_DATA(val);
 	}
 }
 
@@ -605,22 +738,22 @@ static int obj_setitem(PyTdbObject *self, PyObject *key, PyObject *value)
 	TDB_DATA tkey, tval;
 	int ret;
 	PyErr_TDB_RAISE_RETURN_MINUS_1_IF_CLOSED(self);
-	if (!PyString_Check(key)) {
-		PyErr_SetString(PyExc_TypeError, "Expected string as key");
+	if (!PyBytes_Check(key)) {
+		PyErr_SetString(PyExc_TypeError, "Expected bytestring as key");
 		return -1;
 	}
 
-	tkey = PyString_AsTDB_DATA(key);
+	tkey = PyBytes_AsTDB_DATA(key);
 
 	if (value == NULL) { 
 		ret = tdb_delete(self->ctx, tkey);
 	} else { 
-		if (!PyString_Check(value)) {
+		if (!PyBytes_Check(value)) {
 			PyErr_SetString(PyExc_TypeError, "Expected string as value");
 			return -1;
 		}
 
-		tval = PyString_AsTDB_DATA(value);
+		tval = PyBytes_AsTDB_DATA(value);
 
 		ret = tdb_store(self->ctx, tkey, tval, TDB_REPLACE);
 	}
@@ -637,6 +770,9 @@ static PyMappingMethods tdb_object_mapping = {
 	.mp_subscript = (binaryfunc)obj_getitem,
 	.mp_ass_subscript = (objobjargproc)obj_setitem,
 };
+static PySequenceMethods tdb_object_seq = {
+	.sq_contains = (objobjproc)obj_contains,
+};
 static PyTypeObject PyTdb = {
 	.tp_name = "tdb.Tdb",
 	.tp_basicsize = sizeof(PyTdbObject),
@@ -647,56 +783,94 @@ static PyTypeObject PyTdb = {
 	.tp_repr = (reprfunc)tdb_object_repr,
 	.tp_dealloc = (destructor)tdb_object_dealloc,
 	.tp_as_mapping = &tdb_object_mapping,
+	.tp_as_sequence = &tdb_object_seq,
 	.tp_flags = Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_ITER,
-	.tp_iter = (getiterfunc)tdb_object_iter,
+	.tp_iter = PY_DISCARD_FUNC_SIG(getiterfunc,tdb_object_iter),
 };
 
 static PyMethodDef tdb_methods[] = {
-	{ "open", (PyCFunction)py_tdb_open, METH_VARARGS|METH_KEYWORDS, "open(name, hash_size=0, tdb_flags=TDB_DEFAULT, flags=O_RDWR, mode=0600)\n"
-		"Open a TDB file." },
-	{ NULL }
+	{
+		.ml_name  = "open",
+		.ml_meth  = PY_DISCARD_FUNC_SIG(PyCFunction, py_tdb_open),
+		.ml_flags = METH_VARARGS|METH_KEYWORDS,
+		.ml_doc   = "open(name, hash_size=0, tdb_flags=TDB_DEFAULT, "
+			    "flags=O_RDWR, mode=0600)\nOpen a TDB file."
+	},
+	{ .ml_name = NULL }
 };
 
-void inittdb(void);
-void inittdb(void)
+#define MODULE_DOC "simple key-value database that supports multiple writers."
+
+#if PY_MAJOR_VERSION >= 3
+static struct PyModuleDef moduledef = {
+    PyModuleDef_HEAD_INIT,
+    .m_name = "tdb",
+    .m_doc = MODULE_DOC,
+    .m_size = -1,
+    .m_methods = tdb_methods,
+};
+#endif
+
+PyObject* module_init(void);
+PyObject* module_init(void)
 {
 	PyObject *m;
 
 	if (PyType_Ready(&PyTdb) < 0)
-		return;
+		return NULL;
 
 	if (PyType_Ready(&PyTdbIterator) < 0)
-		return;
+		return NULL;
 
-	m = Py_InitModule3("tdb", tdb_methods,
-		"simple key-value database that supports multiple writers.");
+#if PY_MAJOR_VERSION >= 3
+	m = PyModule_Create(&moduledef);
+#else
+	m = Py_InitModule3("tdb", tdb_methods, MODULE_DOC);
+#endif
 	if (m == NULL)
-		return;
+		return NULL;
 
-	PyModule_AddObject(m, "REPLACE", PyInt_FromLong(TDB_REPLACE));
-	PyModule_AddObject(m, "INSERT", PyInt_FromLong(TDB_INSERT));
-	PyModule_AddObject(m, "MODIFY", PyInt_FromLong(TDB_MODIFY));
+	PyModule_AddIntConstant(m, "REPLACE", TDB_REPLACE);
+	PyModule_AddIntConstant(m, "INSERT", TDB_INSERT);
+	PyModule_AddIntConstant(m, "MODIFY", TDB_MODIFY);
 
-	PyModule_AddObject(m, "DEFAULT", PyInt_FromLong(TDB_DEFAULT));
-	PyModule_AddObject(m, "CLEAR_IF_FIRST", PyInt_FromLong(TDB_CLEAR_IF_FIRST));
-	PyModule_AddObject(m, "INTERNAL", PyInt_FromLong(TDB_INTERNAL));
-	PyModule_AddObject(m, "NOLOCK", PyInt_FromLong(TDB_NOLOCK));
-	PyModule_AddObject(m, "NOMMAP", PyInt_FromLong(TDB_NOMMAP));
-	PyModule_AddObject(m, "CONVERT", PyInt_FromLong(TDB_CONVERT));
-	PyModule_AddObject(m, "BIGENDIAN", PyInt_FromLong(TDB_BIGENDIAN));
-	PyModule_AddObject(m, "NOSYNC", PyInt_FromLong(TDB_NOSYNC));
-	PyModule_AddObject(m, "SEQNUM", PyInt_FromLong(TDB_SEQNUM));
-	PyModule_AddObject(m, "VOLATILE", PyInt_FromLong(TDB_VOLATILE));
-	PyModule_AddObject(m, "ALLOW_NESTING", PyInt_FromLong(TDB_ALLOW_NESTING));
-	PyModule_AddObject(m, "DISALLOW_NESTING", PyInt_FromLong(TDB_DISALLOW_NESTING));
-	PyModule_AddObject(m, "INCOMPATIBLE_HASH", PyInt_FromLong(TDB_INCOMPATIBLE_HASH));
+	PyModule_AddIntConstant(m, "DEFAULT", TDB_DEFAULT);
+	PyModule_AddIntConstant(m, "CLEAR_IF_FIRST", TDB_CLEAR_IF_FIRST);
+	PyModule_AddIntConstant(m, "INTERNAL", TDB_INTERNAL);
+	PyModule_AddIntConstant(m, "NOLOCK", TDB_NOLOCK);
+	PyModule_AddIntConstant(m, "NOMMAP", TDB_NOMMAP);
+	PyModule_AddIntConstant(m, "CONVERT", TDB_CONVERT);
+	PyModule_AddIntConstant(m, "BIGENDIAN", TDB_BIGENDIAN);
+	PyModule_AddIntConstant(m, "NOSYNC", TDB_NOSYNC);
+	PyModule_AddIntConstant(m, "SEQNUM", TDB_SEQNUM);
+	PyModule_AddIntConstant(m, "VOLATILE", TDB_VOLATILE);
+	PyModule_AddIntConstant(m, "ALLOW_NESTING", TDB_ALLOW_NESTING);
+	PyModule_AddIntConstant(m, "DISALLOW_NESTING", TDB_DISALLOW_NESTING);
+	PyModule_AddIntConstant(m, "INCOMPATIBLE_HASH", TDB_INCOMPATIBLE_HASH);
 
-	PyModule_AddObject(m, "__docformat__", PyString_FromString("restructuredText"));
+	PyModule_AddStringConstant(m, "__docformat__", "restructuredText");
 
-	PyModule_AddObject(m, "__version__", PyString_FromString(PACKAGE_VERSION));
+	PyModule_AddStringConstant(m, "__version__", PACKAGE_VERSION);
 
 	Py_INCREF(&PyTdb);
 	PyModule_AddObject(m, "Tdb", (PyObject *)&PyTdb);
 
 	Py_INCREF(&PyTdbIterator);
+
+    return m;
 }
+
+
+#if PY_MAJOR_VERSION >= 3
+PyMODINIT_FUNC PyInit_tdb(void);
+PyMODINIT_FUNC PyInit_tdb(void)
+{
+    return module_init();
+}
+#else
+void inittdb(void);
+void inittdb(void)
+{
+    module_init();
+}
+#endif
